#include"swr_mesh.hpp"
#include<cstdio>




namespace swr{




Mesh::
Mesh():
vertex_count(0)
{
}


Mesh::
Mesh(const Mesh&  rhs)
{
  *this = rhs;
}




void
Mesh::
clear()
{
  freed_indexes.clear();

  vertexes.clear();

  edge_list.clear();

  vertex_count = 0;

  matrix.clear();
}


Mesh&
Mesh::
operator=(const Mesh&  rhs)
{
  clear();

  vertexes      = rhs.vertexes;
  vertex_count  = rhs.vertex_count;
  freed_indexes = rhs.freed_indexes;
  edge_list     = rhs.edge_list;
  matrix        = rhs.matrix;

  return *this;
}


const AdjacencyMatrix&
Mesh::
get_matrix() const
{
  return matrix;
}


uint32_t
Mesh::
get_vertex_count() const
{
  return vertex_count;
}


uint32_t
Mesh::
get_edge_count() const
{
  return edge_list.size();
}


Mesh::index_t
Mesh::
insert_vertex(Point*  pt, bool  check_repeat)
{
    if(!pt)
    {
      return -1;
    }


  int  i;

    if(check_repeat)
    {
      i = find_vertex(*pt);

        if(i >= 0)
        {
          return i;
        }
    }


    if((vertex_count+1) >= matrix.get_size())
    {
      matrix.resize(vertex_count*2);
    }


  Vertex*  v;

    if(freed_indexes.size())
    {
      i = freed_indexes.back();

      freed_indexes.pop_back();

      v = &vertexes[i];
    }

  else
    {
      i = vertexes.size();

      vertexes.emplace_back();

      v = &vertexes.back();
    }


  v->reset(pt,i);

  ++vertex_count;

  return i;
}


Mesh::index_t
Mesh::
find_vertex(const Point&  pt) const
{
  int  i = 0;

    for(auto&  v: vertexes)
    {
        if(v && (*v.get_point() == pt))
        {
          return i;
        }


      ++i;
    }


  return -1;
}


int
Mesh::
make_edge(Point*  a, Point*  b, bool  check_repeat)
{
  auto  ai = insert_vertex(a,check_repeat);
  auto  bi = insert_vertex(b,check_repeat);

  int  flag = 0;

    if(ai >= 0){flag |= 1;}
    if(bi >= 0){flag |= 2;}

  flag |= make_edge(ai,bi);

  return flag;
}


int
Mesh::
make_edge(index_t  a, index_t  b, bool  check_repeat)
{
    if((a < 0) ||
       (b < 0))
    {
      return 0;
    }


    if(a < b)
    {
      std::swap(a,b);
    }


    if(check_repeat)
    {
        for(auto&  e: edge_list)
        {
            if(e.test(a,b))
            {
              return 8;
            }
        }
    }


    if(!test_edge(a,b))
    {
      vertexes[a].increase_count();
      vertexes[b].increase_count();

      vertexes[a].append_paired_index(b);
      vertexes[b].append_paired_index(a);

      edge_list.emplace_back(a,b);

      matrix.set_value(a,b,1);

      return 4;
    }


  return 0;
}




bool
Mesh::
test_edge(index_t  a, index_t  b) const
{
  return matrix.get_value(a,b);
}




void
Mesh::
print() const
{
    for(auto&  e: edge_list)
    {
      auto&  a = *vertexes[e.a].get_point();
      auto&  b = *vertexes[e.b].get_point();

      printf("mesh.make_edge(Point(%4d,%4d,%4d),Point(%4d,%4d,%4d),true);\n",
        a.x,a.y,a.z,
        b.x,b.y,b.z);
    }
}


std::vector<Mesh::index_t>
Mesh::
collect_vertex_that_adjoins(index_t  i) const
{
  std::vector<index_t>  buf;

     for(auto  ii: vertexes[i].get_paired_index_list())
     {
       buf.emplace_back(ii);
     }


  return std::move(buf);
}


int
Mesh::
count_triangle_that_includes(index_t  a_index, index_t  b_index, index_t*  t) const
{
  auto  a_res = collect_vertex_that_adjoins(a_index);
  auto  b_res = collect_vertex_that_adjoins(b_index);

  int  n = 0;

  index_t  t_index;

    for(auto  a_val: a_res){
    for(auto  b_val: b_res){
        if(a_val == b_val)
        {
          t_index = a_val;

          ++n;
        }
    }}


    if(t)
    {
      *t = t_index;
    }


  return n;
}


void
Mesh::
unrefer_vertex(index_t  i)
{
  auto&  v = vertexes[i];

    if(v.get_count())
    {
      v.decrease_count();
    }


    if(!v.get_count())
    {
      v.reset();

      freed_indexes.emplace_back(i);

      --vertex_count;
    }
}


void
Mesh::
remove_edge(index_t  a, index_t  b)
{
    if(a < b)
    {
      std::swap(a,b);
    }


    if(matrix.set_value(a,b,0))
    {
      vertexes[a].remove_paired_index(b);
      vertexes[b].remove_paired_index(a);

      auto   it = edge_list.begin();
      auto  end = edge_list.end();

        while(it != end)
        {
            if(it->test(a,b))
            {
              edge_list.erase(it);
              break;
            }


          ++it;
        }


      unrefer_vertex(a);
      unrefer_vertex(b);
    }
}


void
Mesh::
remove(index_t  a, index_t  b, index_t  c)
{
  bool  a_res = (count_triangle_that_includes(a,b) <= 1);
  bool  b_res = (count_triangle_that_includes(b,c) <= 1);
  bool  c_res = (count_triangle_that_includes(a,c) <= 1);

    if(a_res){remove_edge(a,b);}
    if(b_res){remove_edge(b,c);}
    if(c_res){remove_edge(a,c);}
}


bool
Mesh::
cut_polygon(const Color&  color, std::list<Polygon>&  dst)
{
    if(matrix.get_count() >= 3)
    {
        for(auto&  e: edge_list)
        {
          index_t  t;

          auto  n = count_triangle_that_includes(e.a,e.b,&t);

            if(n == 1)
            {
              dst.emplace_back(color,*vertexes[e.a].get_point(),
                                     *vertexes[e.b].get_point(),
                                     *vertexes[  t].get_point());

              remove(e.a,e.b,t);

              return true;
            }
        }
    }


  return false;
}


void
Mesh::
convert_to_lines(const Color&  color, std::list<Line>&  lines) const
{
    for(auto&  e: edge_list)
    {
      auto&  a = *vertexes[e.a].get_point();
      auto&  b = *vertexes[e.b].get_point();

      lines.emplace_back(color,a,b);
    }
}



}




