#include"swr_object.hpp"
#include<cmath>
#include<algorithm>
#include<cstdlib>




namespace swr{


Object::
Object():
parent(nullptr)
{
}


Object::
Object(Object&&  rhs)
{
  *this = std::move(rhs);
}




void
Object::
clear()
{
  polygons.clear();
     lines.clear();
  children.clear();
}


Object&
Object::
operator=(Object&&  rhs)
{
  parent = nullptr;

  center = rhs.center;
  x_degree = rhs.x_degree;
  y_degree = rhs.y_degree;
  z_degree = rhs.z_degree;

  name     = std::move(rhs.name    );
  polygons = std::move(rhs.polygons);
  lines    = std::move(rhs.lines   );
  children = std::move(rhs.children);

  return *this;
}


Object&
Object::
make_child()
{
  children.emplace_back(Object());

  auto&  obj = children.back();

  obj.parent = this;

  return obj;
}


Object
Object::
transform(const Ro*  parent_ro, const Ro*  view_ro, const Vector*  light) const
{
  Object  obj;

  Ro  ro = *this;

    if(parent)
    {
      ro.center += parent_ro->center;

      parent_ro->rotate(ro.center);

      ro.x_degree += parent_ro->x_degree;
      ro.y_degree += parent_ro->y_degree;
      ro.z_degree += parent_ro->z_degree;
    }


    for(auto&  poly: polygons)
    {
      auto  tmp = poly;

      tmp.a_point += ro.center;
      tmp.b_point += ro.center;
      tmp.c_point += ro.center;

      ro.rotate(tmp.a_point);
      ro.rotate(tmp.b_point);
      ro.rotate(tmp.c_point);

        if(light)
        {
          auto  v = tmp.get_vector();

          v.normalize();

          auto  d = Vector::dot_product(v,*light);

          auto  degree = std::abs(to_degree(std::cos(d)));

            if(degree <= 90)
            {
              constexpr int  unit = (255<<16)/90;

              degree = 90-degree;

              tmp.color.r = ((unit*degree)/255*tmp.color.r)>>16;
              tmp.color.g = ((unit*degree)/255*tmp.color.g)>>16;
              tmp.color.b = ((unit*degree)/255*tmp.color.b)>>16;
            }

          else
            {
              tmp.color.r = 0;
              tmp.color.g = 0;
              tmp.color.b = 0;
            }
        }


        if(view_ro)
        {
          tmp.a_point -= view_ro->center;
          tmp.b_point -= view_ro->center;
          tmp.c_point -= view_ro->center;

          view_ro->rotate(tmp.a_point);
          view_ro->rotate(tmp.b_point);
          view_ro->rotate(tmp.c_point);
        }


      tmp.sort();

      obj.polygons.emplace_back(tmp);
    }


    for(auto&  ln: lines)
    {
      auto  tmp = ln;

      tmp.a += ro.center;
      tmp.b += ro.center;

      ro.rotate(tmp.a);
      ro.rotate(tmp.b);

        if(view_ro)
        {
          tmp.a -= view_ro->center;
          tmp.b -= view_ro->center;

          view_ro->rotate(tmp.a);
          view_ro->rotate(tmp.b);
        }


      obj.lines.emplace_back(tmp);
    }


    for(auto&  child: children)
    {
      obj.children.emplace_back(child.transform(&ro,view_ro,light));
    }


  return std::move(obj);
}


void
Object::
render(Screen&  scr) const
{
    for(auto&  poly: polygons)
    {
      poly.render(scr);
    }


    for(auto&  ln: lines)
    {
      ln.render(scr);
    }


    for(auto&  child: children)
    {
      child.render(scr);
    }
}




void
Object::
push_line(const Color&  color, const Point&  a, const Point&  b)
{
  lines.emplace_back(color,a,b);
}


void
Object::
push_polygon(const Color&  color, const Point&  a, const Point&  b, const Point&  c)
{
  polygons.emplace_back(color,a,b,c);
}




}




